#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <EzGraph.h>
#include "config.h"
#include "layout.h"
#include "console.h"
#include "command.h"
#include "kana_spelling.h"
#include "message.h"
#include "verb.h"
#include "noun.h"
#include "history.h"
#include "map.h"
#include "graphic.h"

typedef enum {
	ct_none, ct_1b, ct_2b
} char_type_t;

static char text[CONSOLE_BUFFER_LINES][CONSOLE_CHR_WIDTH+1];
static int tail = 0;
static int rollback = 0;
static char input_line[CONSOLE_CHR_WIDTH+1] = DEFAULT_PROMPT;
static char_type_t char_type[CONSOLE_CHR_WIDTH+1];
static int input_len = PROMPT_LEN;
static input_mode_t input_mode;
//static int shift_down, control_down;
static int space_count = 0;
static int noun_head = 0;
static int roman_head = 0; // ޻Ѵо1ХʸƬ(Ѵ)

static void (*wait_return_callback)(void) = NULL;

void init_console(void){
	EzSetKeyHandler(console_keyh);
	EzSetKeyReleaseHandler(console_key_releaseh);
//    kana_spelling_init(input_line + PROMPT_LEN, CONSOLE_CHR_WIDTH-3, CURSOR);
    kana_spelling_init();
    clear_input_line();
}

void clear_console(void){
    int i;
    for (i = 0 ; i < CONSOLE_BUFFER_LINES ; i++){
        text[i][0] = '\0';
    }
    tail = 0;
	rollback = 0;
    clear_input_line();
//	shift_down = 0;
//	control_down = 0;
}

void clear_input_line(void){
    input_len = PROMPT_LEN;
	roman_head = input_len;
	space_count = 0;
	noun_head = 0;
	set_history_mode(hm_verb);
    input_line[input_len] = CURSOR;
    input_line[input_len+1] = '\0';
//        kana_spelling_clear_buf();
}
void clear_input_noun(void){
    input_len = noun_head;
	roman_head = input_len;
    input_line[input_len] = CURSOR;
    input_line[input_len+1] = '\0';
// kanaط
}
void erase_cursor(void){
    int len = strlen(input_line);
    if (input_line[len-1] == CURSOR){
        // Υʸ򥫥å
        input_line[len-1] = '\0';
    }
//        kana_spelling_erase_cursor();
}
void append_input_line(char *str){
    if (str == NULL) str = "";
	erase_cursor();
    strcat(input_line + strlen(input_line), str);
    input_len = strlen(input_line);
	roman_head = input_len;
    input_line[input_len] = CURSOR;
    input_line[input_len+1] = '\0';
//        kana_spelling_append_input_line(str);
}
void console_key_releaseh(int key){
/*	switch (key) {
	case EZ_Control_L:
	case EZ_Control_R:
        control_down = 0;
		break;
	case EZ_Shift_L:
	case EZ_Shift_R:
        shift_down = 0;
		break;
	default:
		break;
	}
*/
}
void console_keyh(int key){
	char *kana;
	int roman_len, kana_len;
	int i;

	switch (key) {
/*
	case EZ_Control_L:
	case EZ_Control_R:
        control_down = 1;
		break;
	case EZ_Shift_L:
	case EZ_Shift_R:
        shift_down = 1;
		break;
*/
	case 8: // BS
        if (input_len > PROMPT_LEN){
            input_len --;
			if (char_type[input_len] == ct_2b){
				input_len --;
				roman_head -= 2;
			}
			if (input_line[input_len] == ' '){
				space_count --;
				roman_head --;
				printf("space count:%d\n", space_count);
				if (space_count == 0){
					set_history_mode(hm_verb);
					noun_head = 0;
				}
			}
			input_line[input_len] = CURSOR;
			input_line[input_len+1] = '\0';
        }
//            kana_spelling_backspace();
		break;
	case 13:
		if (rollback > 0){
			rollback = 0;
		}
		else {
			erase_cursor();
			put_console(input_line);
		    interpret_line(input_line+PROMPT_LEN);
			if (wait_return_callback == NULL){
				clear_input_line();
	        }
		}
		break;
	case EZ_F1:
		set_config_input((get_config_input() +1) % 3);
		printf("change input mode to %d\n", input_mode);
		break;
	case EZ_F2:
        toggle_show_minimap();
        draw_graphic();
        break;
	case EZ_F3:
        toggle_minimap_location();
        draw_graphic();
        break;
	case EZ_F4:
        toggle_overlap_minimap();
        draw_graphic();
        break;
	case EZ_Page_Up:
	case 'u'-'a'+1:
		if (rollback+CONSOLE_CHR_HEIGHT < CONSOLE_BUFFER_LINES){
			rollback ++;
		}
		break;
	case EZ_Page_Down:
	case 'd'-'a'+1:
		if (rollback > 0){
			rollback --;
		}
		break;
	case EZ_Up:
	case 'p'-'a'+1:
		prev_history_line(input_line);
		if (space_count == 0){
			clear_input_line();
		}
		else {
			clear_input_noun();
		}
		append_input_line(get_history_line(input_line));
		break;
	case EZ_Down:
	case 'n'-'a'+1:
		next_history_line(input_line);
		if (space_count == 0){
			clear_input_line();
		}
		else {
			clear_input_noun();
		}
		append_input_line(get_history_line(input_line));
		break;
 	case 'c'-'a'+1:
		append_input_line("^C");
		erase_cursor();
		put_console(input_line);
		clear_input_line();
		break;
	default:
        // Windowsctype0-255ʳͤassert顼ˤʤؤн
        // ɳü쥳ɤ̵
        // (ɤΤߤ󥽡ϤǤʤʸʤΤ)
//        printf("key:%x(%c) CTRL:%d\n", key, key, control_down);
        if (key & 0xffffff00) return;
		if (key == ' '){
			if (space_count == 0){
				set_history_mode(hm_noun);
				noun_head = input_len+1;
			}
			roman_head ++;
            space_count ++;
			printf("space count:%d\n", space_count);
		}
		if (input_len < CONSOLE_CHR_WIDTH-1){
  			key = tolower(key);
            if (isalnum(key) || key == ' ' || key == '-'){
				char_type[input_len] = ct_1b;
                input_line[input_len] = key;
				input_len ++;
                input_line[input_len] = CURSOR;
                input_line[input_len+1] = '\0';
            }
			if (input_mode != im_english){
				roman_len = input_len-roman_head;
				kana = kana_spelling_conv(&input_line[roman_head], &roman_len, &kana_len);
				if (kana){
//					kana_len = strlen(kana);
					printf("kana:%s:%d:%d\n", kana, roman_len, kana_len);
					for (i = 0 ; i < roman_len ; i++){
						char_type[roman_head + kana_len + i] = ct_1b;
						input_line[roman_head + kana_len + i] = input_line[input_len - roman_len];
					}
					for (i = 0 ; i < kana_len ; i++){
						char_type[roman_head + i] = ct_2b;
						input_line[roman_head + i] = kana[i];
					}
					input_len = roman_head + roman_len + kana_len;
					roman_head += kana_len;
	                input_line[input_len] = CURSOR;
		            input_line[input_len+1] = '\0';
				}

			}
//                if (isalnum(key) || key == ' ' || key == '-'){
//                    kana_spelling_input_key(key);
//                }
		}
	}
	printf("input_len:%d roman_head:%d\n", input_len, roman_head);
	draw_console();
}

// !!! Ȥꤢ2ХȤιƬ§ʸѤߡ
// !!! Զ礬ʤ1Хȶ§1Хȱñʬ뤫ϸǹͤ
void put_console(char *line){
	// reserveΤϴܥ󥽡ʬ§δط;͵򸫤Ƥ
    static unsigned char reserve[CONSOLE_CHR_WIDTH*2+1] = "";
    static int reserve_len = 0;
    // ʸɽΤʤǰƤ֤ȤˤϤchar*ϤΤǡ
    // ǥ㥹ȤƤ
    unsigned char *str = (unsigned char *)line;
	static unsigned char *nohead_2b
        = (unsigned char *)NOHEAD_2B_MARKS NOHEAD_2B_KANA;
	static unsigned char *notail_2b
        = (unsigned char *)NOTAIL_2B_MARKS;
	int len = 0;
	int head_ok = -1, tail_ok = -1;
	unsigned char *sep = str, *p = str, *prev_p = NULL;
	unsigned char *chk;
	int chr2b1 = 0;
    int text_offset = 0;

    // η³ԤФޤ򥳥ԡ
	// Ĥäʬϡ̤̤ΤϤ
    if (reserve_len != 0){
        memcpy(text[tail], reserve, reserve_len);
        text_offset = reserve_len;
        len = reserve_len;
    }

	while (*p != '\0'){
		// 2Хʸ1ХܤΤȤ
		if (Is2bChr1(*p)){
			chr2b1 = *p;
			prev_p = p;
			len ++;
			p ++;
//		}
		// 2Хʸ2ХܤΤȤ
//		else if (chr2b1){
			// 2Хȹ§ʸ˳뤫å
			chk = nohead_2b;
			head_ok = -1;
			while (*chk != '\0'){
				if (chr2b1 == chk[0] && *p == chk[1]){
					head_ok = 0;
					break;
				}
				chk += 2;
			}
			// ʸtail_okĤʸhead_okʤ顢ξʸ֤ʬǽ
			if (tail_ok && head_ok){
				sep = prev_p;
			}
			// ʬǽåΤᡢʸ2ХȹƬ§ʸ˳뤫å
			chk = notail_2b;
			tail_ok = -1;
			while (*chk != '\0'){
				if (chr2b1 == chk[0] && *p == chk[1]){
					tail_ok = 0;
					break;
				}
				chk += 2;
			}
			chr2b1 = 0;
			len ++;
			p ++;
		}
		// 1ХʸΤȤ
		else {
			assert(chr2b1 == 0);
		    // '\n'и鶯
			if (*p == '\n'){
		        memcpy(text[tail]+text_offset, str, len-text_offset);
				text[tail][len] = '\0';
			    tail = (tail + 1) % (CONSOLE_BUFFER_LINES);
	            text_offset = 0;
				len = 0;
				prev_p = NULL;
				p++;
				str = p;
				sep = str;
				continue;
//				reserve_len = 0;

			}
			// !! IsAlphaϢ³֤sepˤʤ褦ˤơñ줬ڤʤ褦ˤɲ
			prev_p = p;
			sep = prev_p;
			len ++;
			p ++;
		}
		if (len > CONSOLE_CHR_WIDTH){
			len = sep - str;
			memcpy(text[tail]+text_offset, str, len);
	        text[tail][text_offset + len] = '\0';
		    tail = (tail + 1) % (CONSOLE_BUFFER_LINES);
            text_offset = 0;
			len = 0;
			str = sep;
			// λĤϰܤιԽϤǽϤϤ
			reserve_len = 0;
            // Ƭ˥ڡȤϽϤʤ
            while (*str == ' '){
                str++;
            }
			p = str;
		}
	}
    if (*(p-1) == '\\'){
		// η³ԤĤäƤʤȤ
		if (reserve_len == 0){
			// ߤνϻĤ³ԤȤreserve
	        memcpy(reserve, str, len);
		    reserve_len = len-1; // ³'\'ʬ򸺤餷Ƥ
		}
		// η³ԤĤäƤʤС
		// 1ԤϤʤˤޤ³˽вäƤޤäȤ
		else {
	        memcpy(reserve + reserve_len, str, len);
		    reserve_len = len-1; // ³'\'ʬ򸺤餷Ƥ
		}
    }
	else {
        memcpy(text[tail]+text_offset, str, len-text_offset);
        text[tail][len] = '\0';
        tail = (tail + 1) % (CONSOLE_BUFFER_LINES);
		reserve_len = 0;
    }
    draw_console();
}


void draw_console(void){
    int i, ix;
    ix = (tail + (CONSOLE_BUFFER_LINES-CONSOLE_CHR_HEIGHT-rollback) + 1) % CONSOLE_BUFFER_LINES;
    EzSetColorByRGB(0, 0, 0);
    //    EzSetColorByRGB(255, 0, 0);
    EzFillBox(CONSOLE_LT_X-1, CONSOLE_LT_Y-1,
			  CONSOLE_RB_X+1, CONSOLE_RB_Y+1);
    EzSetColorByRGB(255, 255, 255);
    EzDrawBox(CONSOLE_LT_X-2, CONSOLE_LT_Y-2,
		      CONSOLE_RB_X+2, CONSOLE_RB_Y+2);
//	EzDrawBox(CONSOLE_LT_X, CONSOLE_LT_Y,
//		      CONSOLE_LT_X+CONSOLE_WIDTH*FONT_SIZE,
//			  CONSOLE_LT_Y+CONSOLE_HEIGHT*FONT_SIZE);
    for (i = 0 ; i < CONSOLE_CHR_HEIGHT-1 ; i++){
        //       printf("%d %d %s\n", i, ix, text[ix]);
        EzDrawString(CONSOLE_LT_X, CONSOLE_LT_Y+i*LINE_SKIP+LINE_SKIP, text[ix]);       
        ix = (ix + 1) % CONSOLE_BUFFER_LINES;
    }
//    EzSetColorByRGB(255, 255, 0);
	EzDrawString(CONSOLE_LT_X, CONSOLE_LT_Y+CONSOLE_CHR_HEIGHT*LINE_SKIP, input_line);
}

void set_console_input(input_mode_t mode){
    input_mode = mode;
    if (mode == im_hiragana) kana_spelling_set_hiragana();
    else if (mode == im_katakana) kana_spelling_set_katakana();
}

void set_console_prompt(char *prompt){
    //    printf("set_console_prompt\n");    
    if (prompt == NULL){
        strcpy(input_line, DEFAULT_PROMPT);
    }
    else {
        strcpy(input_line, prompt);
    }
}

void set_wait_return_callback(void (*callback)(void)){
     printf("set_wait_return_callback called\n");
	EzSetKeyHandler(wait_return_keyh);
    strcpy(input_line, "[꥿󥭡򲡤Ƥ]");
    wait_return_callback = callback;
}


void wait_return_keyh(int key){
    void (*callback)(void);
    printf("wait return keyh\n");
    assert(wait_return_callback != NULL);  // callback==NULLʤ餳ΥϥɥϸƤФʤϤ
	if (key == 13){
        EzSetKeyHandler(console_keyh);
        callback = wait_return_callback;
        wait_return_callback = NULL;
        set_console_prompt(DEFAULT_PROMPT);
        clear_input_line();
        callback();
    }
    draw_console();
	// Хåؿǥե饰ѹƤ
	// եåѲ뤫⤷ʤΤǺ
	draw_graphic();
}

void strip_space(char *line){
    char *p = line, *q;
    while (*p != '\0'){
        if (Is2bChr1(*p)){
            p += 2;
            continue;
        }
        if (*p != ' '){
            p++;
            continue;
        }
        q = p;
        while (*(q+1) != '\0'){
            *q = *(q+1);
            q++;
        }
        *q = '\0';
    }
}
