#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "config.h"
#include "status.h"
#include "noun.h"
#include "uty.h"
#include "message.h"
#include "graphic.h"

static status_t status;
static char *exe_dir;
static unsigned char major_version = MAJOR_VERSION;
static unsigned char minor_version = MINOR_VERSION;
#define GetVersion(major, minor) (major*100+minor)


input_mode_t input_mode;

flag_sequence_t flag_sequence_dragon = {
    {//fi_dragon_library,
     fi_dragon_corridor_2f_r,
     fi_dragon_guestroom_1f,
     fi_dragon_shotdown,
     fi_invalid},
    0, 1
};

flag_sequence_t *flag_sequence_table[] = {
    &flag_sequence_dragon,
};

#define FLAG_SEQUENCE_NUM (sizeof(flag_sequence_table)/sizeof(flag_sequence_t*))

point_t normal_points[] = {
    {fi_leaflet_taken, 1, 1},
    {fi_entered_house, 1, 5},
    {fi_bathtub_empty, 1, 5},
    {fi_mirror_taken, 1, 5},
    {fi_glasses_taken, 1, 5},
};
point_t maniac_points[] = {
    {fi_medal_taken,     1, 1},
    {fi_mailbox_has_closed,     1, 1},
    {fi_toilet_used, 1, 1},
    {fi_phone_call_answered, 1, 1},
};

#define MANIAC_POINT_NUM (sizeof(maniac_points)/sizeof(point_t))
#define NORMAL_POINT_NUM (sizeof(normal_points)/sizeof(point_t))


void status_init(void){
    int i, j;
    set_config_input(DEFAULT_INPUT_MODE);
    // ƥե饰󥹤Υե饰num˥åȤ
    for (i = 0 ; i < FLAG_SEQUENCE_NUM ; i++){
        for (j = 0 ; j < FLAG_SEQUENCE_MAXLEN ; j++){
            //            printf("%d %d %d\n", i, j, flag_sequence_table[i]->flags[j]);
            if (flag_sequence_table[i]->flags[j] == fi_invalid){
                flag_sequence_table[i]->num = j;
                break;
            }
        }
        assert(j < FLAG_SEQUENCE_MAXNUM);
    }
 }
void clear_status(void){
    int i;
    flag_sequence_id_t fs;

    status.current_area = ai_title;
    status.item_num = 0;
    for (fs = fs_begin ; fs < fs_end ; fs++){
        clear_flag_sequence(fs);
    }
    for (i = 0 ; i < FLAG_CHAR_NUM ; i++){
        status.flags[i] = 0;
    }
}


area_id_t current_area_id(void){
    return status.current_area;
}

void set_area_id(area_id_t id){
    status.current_area = id;
}

int is_flag_on(flag_id_t fid){
    return (((status.flags[fid/8] >> fid%8) & 1) == 1);
}

int is_flag_off(flag_id_t fid){
    return (((status.flags[fid/8] >> fid%8) & 1) == 0);
}

void set_flag(flag_id_t fid){
    status.flags[fid/8] |= 1 << fid%8;
    //    draw_graphic();
}
void reset_flag(flag_id_t fid){
	if (is_flag_on(fid)){
        status.flags[fid/8] -=  1 << fid%8;
	}
    //    draw_graphic();
}
void toggle_flag(flag_id_t fid){
	if (is_flag_on(fid)){
		reset_flag(fid);
	}
	else {
		set_flag(fid);
	}
//    status.flags[fid] = !status.flags[fid];
    //    draw_graphic();
}
void test_and_set_flag(flag_id_t fid, char *msg_success, char *msg_failure){
    if (is_flag_off(fid)){
        set_flag(fid);
        if (msg_success){
            message(msg_success);
        }
    }
    else {
        if (msg_failure){
            message(msg_failure);
        }
    }
}
void test_and_reset_flag(flag_id_t fid, char *msg_success, char *msg_failure){
    if (is_flag_on(fid)){
        reset_flag(fid);
        message(msg_success);
    }
    else {
        message(msg_failure);
    }
}

int has_item(noun_id_t nid){
	int i;

	for (i = 0 ; i < status.item_num ; i++){
		if (status.items[i] == nid) return 1;
	}
    return 0;
}

void get_item(noun_id_t nid){
	status.items[status.item_num++] = nid;
    draw_graphic();
}
void lose_item(noun_id_t nid){
	int i, j;

	for (i = 0 ; i < status.item_num ; i++){
		if (status.items[i] == nid){
			for (j = i ; j < status.item_num-1 ; j++){
				status.items[j] = status.items[j+1];
			}
			status.item_num --;
			return;
		}
	}
}

int get_item_num(void){
	return status.item_num;
}
noun_id_t get_item_id(int ix){
	return status.items[ix];
}


void show_score(char *msg, point_t points[], int point_num){
    int i, score = 0, total_score = 0;
    
    for (i = 0 ; i < point_num ; i++){
        total_score += points[i].score;
        if (status.flags[points[i].id] == points[i].value){
            score += points[i].score;
        }
    }
    message("%s:%d/%d", msg, score, total_score);
}
void show_normal_score(void){
    show_score("ߤ", normal_points, NORMAL_POINT_NUM);
}
void show_maniac_score(void){
    show_score("ޥ˥ɾ", maniac_points, MANIAC_POINT_NUM);
}



// SAVE_DIR˥⡼modeǥǡեfile򳫤
// 
FILE *open_data(char *file, char *mode){
    char path[512];
    char *home;

    home = getenv("HOME");
    if (home != NULL){
        sprintf(path, "%s/%s", home, SAVE_DIR);
        if (strchr(mode, 'w')){
            if (check_dir(path) == 0){
                printf("Creating directory '%s'...", path);
                if (make_dir(path) == 0){
                    printf("success.\n");
                }
                else {
                    printf("failure.\n");
                    return NULL;
                }
            }
        }
        sprintf(path, "%s/%s/%s", home, SAVE_DIR, file);
    }
    else {
        sprintf(path, "%s/%s", exe_dir, file);
    }
    return fopen(path, mode);
}

int load_data(void){
    FILE *fp;
    unsigned char file_major_version = 0;
    unsigned char file_minor_version = 0;
    int version = GetVersion(MAJOR_VERSION, MINOR_VERSION);
	int file_version;
	unsigned char id[APPLICATION_ID_SIZE];

    if ((fp = open_data(SAVE_FILE, "rb")) == NULL){
        return -1;
    }
    // إåפʤɤ߹
	fread(id, 1, APPLICATION_ID_SIZE, fp);
	if (memcmp(id, APPLICATION_ID, APPLICATION_ID_SIZE) != 0){
		return -1;
	}
    fread(&file_major_version, sizeof(file_major_version), 1, fp);
    fread(&file_minor_version, sizeof(file_minor_version), 1, fp);
    // ̤ΤΥСʤɤ߹
	file_version = GetVersion(file_major_version, file_minor_version);
    if (file_version <= GetVersion(0, 10) || file_version > version){
        return -1;
    }
    fread(&status, sizeof(status_t), 1, fp);
    fclose(fp);
    return 0;
}

int save_data(void){
    FILE *fp;

    if ((fp = open_data(SAVE_FILE, "wb")) == NULL){
        return -1;
    }
	fwrite(APPLICATION_ID, APPLICATION_ID_SIZE, 1, fp);
    fwrite(&major_version, sizeof(char), 1, fp);
    fwrite(&minor_version, sizeof(char), 1, fp);
    fwrite(&status, sizeof(status_t), 1, fp);
    fclose(fp);
    return 0;
}

void set_config_input(input_mode_t mode){
    input_mode = mode;
    set_console_input(mode);
}
input_mode_t get_config_input(void){
    return input_mode;
}

// ꤷե饰󥹤ηե饰򤹤٤Ʋ
void clear_flag_sequence(flag_sequence_id_t fs){
    int i;
    flag_sequence_t *fseq = flag_sequence_table[fs];
    flag_sequence_status_t *stat = &status.flag_sequence[fs];

    for (i = 0 ; i < fseq->num ; i++){
        reset_flag(fseq->flags[i]);
    }
    stat->phase = 0;
}    
// ꤷե饰󥹤򳫻Ϥ롣
// Ʊ󥹤ʹǤ졢ƬΥե饰ΤΩ
void start_flag_sequence(flag_sequence_id_t fs){
    flag_sequence_t *fseq = flag_sequence_table[fs];

    // ߤοʹԾ˴ؤʤե饰򤤤ä󥯥ꥢ
    clear_flag_sequence(fs);
    set_flag_sequence_active(fs);
    // ǽΥե饰ON
    set_flag(fseq->flags[0]);
    printf("started flag sequence %d\n", fs);
}
// ե饰󥹤ʹԤ
void proceed_flag_sequence(flag_sequence_id_t fs){
    flag_sequence_t *fseq = flag_sequence_table[fs];
    flag_sequence_status_t *stat = &status.flag_sequence[fs];

    printf("%d %d %d\n", fs, stat->phase, fseq->num);
    assert (stat->phase < fseq->num);
    if (fseq->is_exclusive){
        reset_flag(fseq->flags[stat->phase]);
    }
    stat->phase ++;
    set_flag(fseq->flags[stat->phase]);
    printf("proceed flag sequence %d (%d->%d\n",
           fs, stat->phase-1,stat->phase);
}
void abort_flag_sequence(flag_sequence_id_t fs){
    clear_flag_sequence(fs);
    reset_flag_sequence_active(fs);
    printf("end flag sequence %d\n", fs);
}
void set_flag_sequence_active(flag_sequence_id_t fs){
    flag_sequence_status_t *stat = &status.flag_sequence[fs];
    stat->bits |= 1 << FLAG_SEQUENCE_BIT_ACTIVE;
}    
void reset_flag_sequence_active(flag_sequence_id_t fs){
    flag_sequence_status_t *stat = &status.flag_sequence[fs];
    stat->bits &= 0xffff - (1 << FLAG_SEQUENCE_BIT_ACTIVE);
}    
int is_flag_sequence_active(flag_sequence_id_t fs){
    flag_sequence_status_t *stat = &status.flag_sequence[fs];
    return stat->bits & (1 << FLAG_SEQUENCE_BIT_ACTIVE);
}
void set_flag_sequence_complete(flag_sequence_id_t fs){
    flag_sequence_status_t *stat = &status.flag_sequence[fs];
    stat->bits |= 1 << FLAG_SEQUENCE_BIT_COMPLETE;
}    
int is_flag_sequence_complete(flag_sequence_id_t fs){
    flag_sequence_status_t *stat = &status.flag_sequence[fs];
    return stat->bits & (1 << FLAG_SEQUENCE_BIT_COMPLETE);
}

int has_light(void){
    if ((Has(candle) && Is(candle_lit)) ||
        (Has(lantern) && Is(lantern_lit)))
    {
        return 1;
    }
    else {
        return 0;
    }
}
